// Edited by Matthew Ferea and Eric Lam
// a very simple program to read and display TIFF images using fltk and
// libtiff
// written September 19, 1999 by Michael Gleicher

// this program reads a tiff image and displays it into an fltk window
// the goal was to make things as simple as possible

// one thing to notice: tiff reads the image in upside down, and rather than
// write code to flip it over, I instead just display it upside down

// this is what I need to use FlTk
#include <Fl/fl_file_chooser.h>
#include <Fl/Fl.h>
#include <Fl/fl_message.h>
#include <Fl/Fl_Double_Window.h>
#include <Fl/fl_draw.h>

// this is what I need to use libtiff
#include <stdio.h>
#include <tiffio.h>
// this routine reads in a tiff file and stores it in a very simple data structure
int readTiff(char* filename, int& rw, int& rh, unsigned char * &rbits)
{
  // we turn off Warnings (maybe a bad idea) since many photoshop images
  // have newfangled tags that confuse libtiff
  TIFFErrorHandler warn = TIFFSetWarningHandler(0);

  TIFF* tiff = TIFFOpen(filename, "r");

  // turn warnings back on, just in case
  TIFFSetWarningHandler(warn);

  if (tiff) {
    int rc = 1;			// what to return
    uint32 w, h;
    size_t npixels;
    uint32* raster;

    TIFFGetField(tiff, TIFFTAG_IMAGEWIDTH, &w);
    TIFFGetField(tiff, TIFFTAG_IMAGELENGTH, &h);

    // in case my data structures are different
    rw = w;
    rh = h;

    npixels = w * h;
    raster = (uint32*) _TIFFmalloc(npixels * sizeof (uint32));
    if (raster != NULL) {
      if (TIFFReadRGBAImage(tiff, w, h, raster, 0)) {
	rbits = (unsigned char*) raster;
      } else {
	rc=0;
	_TIFFfree(raster);
      }
    } else rc = 0;

    TIFFClose(tiff);
    return rc;
  } else
    return 0;
}

// my very own widget that displays tiff data
// it doesn't do very much, except for displaying the image
// it assumes the image is stored upside down (the way a TIFF image is)
// and is 4 channels (RGBA)
class TiffWidget : public Fl_Widget {
  int w,h;
  unsigned char* image;
  void draw() {
    // notice that we flip the image when we write it to the
    // screen, hence the -w*4. we also have to point at the last
    // row of the image
    unsigned char* imageBottom = image + w * 4 * (h-1);
    fl_draw_image(imageBottom,5,5,w,h,4,-w*4);
  };
  public: 
  TiffWidget(int x, int y, int imageW, int imageH, unsigned char* imageD)
    : Fl_Widget(x,y,imageW,imageH)
    {
      w = imageW;
      h = imageH;
      image = imageD;
    };
};

// simple main to try out these two pieces
int main(int argc, char** argv)
{
  // ask the user for a tiff file to load
  //char* ifilename = "/home/mferea/robocup/ui/ui";
  //char* filename = fl_file_chooser("Pick a TIFF file","{*.tif}",ifilename);
  char * filename = "note.tif";
  // what to read the tiff image into
  int w,h;
  unsigned char* image;

  // read the image
  if (!filename || !readTiff(filename,w,h,image))
    fl_alert("Can't open TIFF file!");
  else {
    Fl::visual(FL_RGB);
    // make a window
    Fl_Window window(500,400,"hello");
    //Fl_Double_Window window(w+10,h+10,filename);
    TiffWidget tw(5,5,w,h,image);
    window.end();
    window.show();
    Fl::run();

    // remember, the image is a TIFF raster, so we need to free it
    _TIFFfree(image);

  }

  return 1;
}
