#ifndef __THREAD_H__
#define __THREAD_H__

#include <iostream>
#include <pthread.h>

    // Borrowed from FLTK
class Mutex
{
    friend class SignalMutex;
    pthread_mutex_t mutex;
    //Mutex(const Mutex&);
    //Mutex& operator=(const Mutex&);
protected:
    Mutex(const pthread_mutexattr_t* a) {pthread_mutex_init(&mutex, a);}
public:
    Mutex() {pthread_mutex_init(&mutex, 0);}
    void lock() {pthread_mutex_lock(&mutex);}
    void unlock() {pthread_mutex_unlock(&mutex);}
    bool trylock() {return pthread_mutex_trylock(&mutex) == 0;}
    ~Mutex() {pthread_mutex_destroy(&mutex);}
};

class SignalMutex : public Mutex
{
    pthread_cond_t cond;
public:
    SignalMutex() : Mutex() {pthread_cond_init(&cond, 0);}
    void signal() {pthread_cond_broadcast(&cond);}
    void signal_one() {pthread_cond_signal(&cond);}
    void wait() {pthread_cond_wait(&cond, &mutex);}
};

class Guard
{
    Mutex& lock;
public:
    Guard(Mutex& m) : lock(m) {lock.lock();}
    Guard(Mutex* m) : lock(*m) {lock.lock();}
    ~Guard() {lock.unlock();}

};

class Thread
{
public:
    bool bKill;

    static void Wait(Thread*);

    Thread();
    void Start();
    void Stop();
    void Resume();
    void Wait();
protected:
    void Run();
    static void* EntryPoint(void*);
    virtual void Setup();
    virtual void Execute();
    virtual void Exit();
private:
    pthread_t m_thread;
    SignalMutex pauseMutex;
    bool bPaused;
};



#endif
