#include "BaseStation.h"

//#define CONNECT_TO_USARSIM false
bool bUT=false;

using namespace std;

BaseStation::BaseStation()
{
    cout << "BaseStation::BaseStation" << endl;

    bStarted = bOK = false;
    gui = new BaseStationGUI(this);
   
    bUT = true;
    if (bUT)
    {
        connectionManager = new ConnectionManager(BASESTATION_NAME,LOCAL_IP_ADDR,"7778");
        joystickManager = new JoystickManager(0);
        slam = new Slam(false);
    }
    else
    {
        connectionManager = new ConnectionManager(BASESTATION_NAME,"localhost","7778");
        joystickManager = new JoystickManager(0);
        slam = new Slam(false);
    }
}

void BaseStation::start()
{
    char s[32];
    Packet* p;
    cout << "sent START" << endl;
    sprintf(s,"START");
    p = new Packet(CMD,strlen(s)+1,s);
    connectionManager->Send(p);
    delete p;
    bStarted = true;
}

void BaseStation::Setup()
{
    cout << "BaseStation::Setup" << endl;
    bOK = false;
    if (connectionManager->SendHandShake(TALON_NAME,"7780"))
        if (connectionManager->WaitForHandShake())
            bOK = true;
    if (bOK)
    {
        //slam->Start();
    }
}

void BaseStation::Execute()
{
    cout << "BaseStation::Execute" << endl;

    if (!bOK)
    {
        cout << "Couldn't connect!" << endl;
        return;
    }
   
    char s[32];
    Packet* cp = NULL;
    Packet* p = NULL;
    int counter = 1;
    TSense sense;
    TOdo odo;
    bool bFirstDataPacket=true;

    //*
    while (!bStarted)
    {
        if (joystickManager->Ready())
        {
            if (joystickManager->AnyButton())
                start();
        }
    }
    //*/
    //start();
    sleep(1);

    do
    {
        cp = p = NULL;
        p = connectionManager->Receive();
        if (p==NULL)
        {
            cout << "null packet p!" << endl;
            break;
        }
        if (p->type==MSG)
        {
            cout << "got MSG: " << p->data << endl;
            float t[181];
            slam->CurrSensation(sense);
            slam->CurrOdometry(odo);
            for (int i=0;i<180;i++)
                t[i] = sense[i].distance/MAP_SCALE;

            gui->scanView->pushLaserRanges(t);
            gui->pushOdoPoint(odo.x,odo.y,odo.theta);
        }
        else if (p->type==DATA)
        {
            cout << "got DATA:" << p->size << endl;
            if (bFirstDataPacket)
            {
                slam->Start();
                bFirstDataPacket = false;
            }
            slam->push(p->data);

            float t[181];
            slam->CurrSensation(sense);
            slam->CurrOdometry(odo);
            for (int i=0;i<180;i++)
                t[i] = sense[i].distance/MAP_SCALE;
            
            gui->scanView->pushLaserRanges(t);
            gui->pushOdoPoint(odo.x,odo.y,odo.theta);
        }
        else if (p->type==IMG)
        {
            cout << "got IMG: " << p->size << endl;

            imgMutex.lock();
            FILE* f = fopen("img.jpg","wb");
            fwrite(p->data,1,p->size,f);
            fclose(f);
            imgMutex.unlock();
            gui->cameraView->redraw();
            fltk::check();
        }
        else
        {
            cout << "got packet with type: " << p->type << " " << p->data << endl;
        }

        //cout << "Signal Strength: " << connectionManager->getSS() << endl;
        //gui->signalStrength->position((double)-connectionManager->getSS());

        if (joystickManager->Ready())
        {
            cp = joystickManager->getPacket();
            if (cp)
            {
                cout << "sent CMD" << endl;
                connectionManager->Send(cp);
                if (strstr(cp->data,"STOP"))
                    counter = 0;
            }
        }
        else
        {
            cout << "sent CMD" << endl;
            //*
            double minL,minR,newspeed,newturnrate;

            minL=1e9;
            minR=1e9;
            for (int j=0;j<90;j++)      if (minR>sense[j].distance/MAP_SCALE) minR=sense[j].distance/MAP_SCALE;
            for (int j=90;j<180;j++)    if (minL>sense[j].distance/MAP_SCALE) minL=sense[j].distance/MAP_SCALE;
            double l=(1e5*minR)/500-100;if (l>100) l=100;
            double r=(1e5*minL)/500-100;if (r>100) r=100;

            newspeed = (r+l)/1e3;
            newturnrate = (r-l);
            newturnrate = MIN(newturnrate,40);
            newturnrate = MAX(newturnrate,-40);

            sprintf(s,"MOVE %.2f %.2f 0.0 0.0 0.0 0.0 0.0:",newspeed,newturnrate);
            //*/
            /*
            sprintf(s,"MOVE .5 30.0 0.0 0.0 0.0 0.0 0.0:");
            //*/
            p = new Packet(CMD,strlen(s)+1,s);
            connectionManager->Send(p);
            delete p;
            p = NULL;
        }
    } while (counter && !bKill);

    {
        sprintf(s,"STOP");
        Packet p(CMD,strlen(s)+1,s);
        connectionManager->Send(&p);
        cout << "sent STOP" << endl;
        usleep(500000);
    }
}

void BaseStation::Exit()
{
    cout << "BaseStation::Exit" << endl;
    connectionManager->Close();
    if (bOK)
    {
        slam->Stop();
        gui->Stop();
        joystickManager->Stop();
    }
}
