#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <syslog.h>

#include <sys/socket.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <gd.h>

#include "Connection.h"

using namespace std;

static const char *progname = "Connection";

void itoa(int i,char *s)
{
    char c, *t;

    if (i < 0) {
        *s++ = '-';
        i = -i; 
    }

    t = s;

    do {
        *t++ = i % 10 + '0';
    } while (i /= 10);

    *t = '\0';

    for (--t; t > s; ++s, --t) {
        c = *s; *s = *t; *t = c;
    }
}


/*
 * Establish a connection to a remote TCP server. First get the list
 * of potential network layer addresses and transport layer port
 * numbers. Iterate through the returned address list until an attempt
 * to establish a TCP connection is successful (or no other
 * alternative exists).
 */

int tcp_connect(char *host, char *port)
{
    struct addrinfo hints, *ai_list, *ai;
    int n, fd = 0;

    memset(&hints, 0, sizeof(hints));
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;

    n = getaddrinfo(host, port, &hints, &ai_list);
    if (n) {
        fprintf(stderr, "%s: getaddrinfo: %s\n",progname,gai_strerror(n));
        exit(EXIT_FAILURE);
    }

    for (ai = ai_list; ai; ai = ai->ai_next) {
        fd = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol);
        if (fd < 0) {
            continue;
        }
        if (connect(fd, ai->ai_addr, ai->ai_addrlen) == 0) {
            break;
        }
        close(fd);
    }

    freeaddrinfo(ai_list);

    if (ai == NULL) {
        fprintf(stderr, "%s: socket or connect: failed for %s port %s\n",
                progname, host, port);
        //exit(EXIT_FAILURE);
        return -1;
    }

    return fd;
}

/*
 * Close a TCP connection. This function trivially calls close() on
 * POSIX systems, but might be more complicated on other systems.
 */

int tcp_close(int fd)
{
    return close(fd);
}

/*
 * Create a listening TCP endpoint. First get the list of potential
 * network layter addresses and transport layer port numbers. Iterate
 * through the returned address list until an attempt to create a
 * listening TCP endpoint is successful (or no other alternative
 * exists).
 */

int tcp_listen(char *port)
{
    struct addrinfo hints, *ai_list, *ai;
    int n, fd = 0, on = 1;
    
    memset(&hints, 0, sizeof(hints));
    hints.ai_flags = AI_PASSIVE;
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;

    n = getaddrinfo(NULL, port, &hints, &ai_list);
    if (n) {
        fprintf(stderr, "%s: getaddrinfo failed: %s\n",
                progname, gai_strerror(n));
        return -1;
    }

    for (ai = ai_list; ai; ai = ai->ai_next) {
        fd = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol);
        if (fd < 0) {
            continue;
        }

        setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));
        if (bind(fd, ai->ai_addr, ai->ai_addrlen) == 0) {
            break;
        }
        close(fd);
    }

    freeaddrinfo(ai_list);

    if (ai == NULL) {
        fprintf(stderr, "%s: bind failed for port %s\n",
                progname, port);
        return -1;
    }

    if (listen(fd, 42) < 0) {
        fprintf(stderr, "%s: listen failed: %s\n",
                progname, strerror(errno));
	close(fd);
        return -1;
    }

    return fd;
}

/*
 * Accept a new TCP connection and write a message about who was
 * accepted to the system log.
 */

int tcp_accept(int listen)
{
    struct sockaddr_storage ss;
    socklen_t ss_len = sizeof(ss);
    char host[NI_MAXHOST];
    char serv[NI_MAXSERV];
    int n, fd;

    fd = accept(listen, (struct sockaddr *) &ss, &ss_len);
    if (fd == -1) {
        syslog(LOG_ERR, "accept failed: %s", strerror(errno));
        return -1;
    }

    n = getnameinfo((struct sockaddr *) &ss, ss_len,
                    host, sizeof(host), serv, sizeof(serv),
                    NI_NUMERICHOST);
    if (n) {
        syslog(LOG_ERR, "getnameinfo failed: %s", gai_strerror(n));
    } else {
        syslog(LOG_DEBUG, "connection from %s:%s", host, serv);
    }

    return fd;
}

Connection::Connection(char* utServerIP,char* utListenPort,char* bot,char* port,char* ip)
{
    strcpy(utHost,utServerIP);
    strcpy(utPort,utListenPort);
    strcpy(hBot,bot);
    strcpy(hPort,port);
    strcpy(hIP,ip);
    excessMax = 512;
    excess = (char *)malloc(excessMax);
}

bool Connection::getOK(int fd)
{
    int n;
    char msg[256];
    n = read(fd,msg,sizeof(msg) - 1);
    msg[n] = '\0';
    char* p = strstr(msg,";");
    if (p) *p = 0;
    cout << msg << endl;
    if (strstr(msg,"OK"))
        return true;
    if (strstr(msg,"already registered"))
    {
        cout << "already registered..." << endl;
        return true;
    }
    return false;
}

float Connection::getSS(char* bot)
{
    char* p;
    int n;
    char msg[256];

    fdc = tcp_connect(utHost,utPort);
    if (fdc < 0)
        return -1;

    strcpy(msg,"GETSS ");strcat(msg,bot);strcat(msg," ");strcat(msg,hBot);strcat(msg,";");
    if (write(fdc,msg,strlen(msg)) != strlen(msg))
        return false;
    //cout << "sent: " << msg << endl;

    while ((n = read(fdc,msg,sizeof(msg) - 1)) > 0)
    {
        msg[n] = '\0';
        p = strstr(msg,";");
        if (p)
        {
            *p = 0;
            cout << msg << endl;
            break;
        }
    }
    tcp_close(fdc);

    if (msg[0]=='O' && msg[1]=='K')
    {
        p = strstr(msg,":");
        *p = 0;
        p++;
        return atof(p);
    }
    return -1;
}

bool Connection::connectNoUT(char* port)
{
    con = tcp_connect("localhost",port);
    //con = tcp_connect("127.0.1.1",port);
    if (con < 0)
        return false;
    return true;
}

bool Connection::listenNoUT()
{
    fdc = tcp_listen(hPort);
    con = tcp_accept(fdc);
    tcp_close(fdc);
    if (con < 0)
        return false;
    return true;
}

bool Connection::registerBot(char* botIP)
{
    char msg[256];
    fdc = tcp_connect(utHost,utPort);
    if (fdc < 0)
        return false;

    strcpy(msg,"register ");strcat(msg,hBot);strcat(msg," ");strcat(msg,botIP);strcat(msg,";");
    cout << "SENT: " << msg << endl;
    if (write(fdc,msg,strlen(msg)) != strlen(msg))
        return false;
    if (getOK(fdc)) 
        return true;
    tcp_close(fdc);
    return false;
}

bool Connection::listen()
{
    int n;
    char msg[256];

    if (!registerBot(hIP))
        return false;

    strcpy(msg,"listen ");strcat(msg,hBot);strcat(msg," ");strcat(msg,hPort);strcat(msg,";");
    if (write(fdc,msg,strlen(msg)) != strlen(msg))
        return false;

    fdl = tcp_listen(hPort);

    if (!getOK(fdc))
        return false;

    //tcp_close(fdc);
    
    con = tcp_accept(fdl);

    return true;
}

bool Connection::connect(char* bot,char* port)
{
    char msg[256];

    if (!registerBot(hIP))
        return false;

    strcpy(msg,"open ");strcat(msg,hBot);strcat(msg," ");strcat(msg,hPort);strcat(msg," ");strcat(msg,bot);strcat(msg," ");strcat(msg,port);strcat(msg,";");
    if (write(fdc,msg,strlen(msg)) != strlen(msg))
        return false;
    cout << "sent: " << msg << endl;
    
    fdl = tcp_listen(hPort);
    
    if (!getOK(fdc))
        return false;
    
    tcp_close(fdc);

    con = tcp_accept(fdl);
    
    return true;
}

bool Connection::send(Packet* p,bool bForUS)
{
    int n;
    char buf[256];
    char typeBuf[8];
    int block = 245;
    int curPos = 0;
    int s = p->size;
    char* data = p->data;

    if (p->type==MSG)   strcpy(typeBuf,"MSG");
    if (p->type==IMG)   strcpy(typeBuf,"IMG");
    if (p->type==CMD)   strcpy(typeBuf,"CMD");
    if (p->type==DATA)  strcpy(typeBuf,"DATA");
   
    if (!bForUS)
    {
        sprintf(buf,"%s %i:",typeBuf,s);
        if (write(con,buf,strlen(buf)) != strlen(buf))
            return false;

        while (curPos < s)
        {
            if ((curPos+block) > s)
                block = s - curPos;
            n = write(con,data+curPos,block);
            if (n != block)
                return false;
            curPos += n;
            //cout << curPos << " of " << s << " bytes transfered" << endl;
        }

       //cout << "transfered " << curPos << " bytes of " << s << endl;
        
       if (curPos != s)
           return false;
    }
    else
    {
        char sbuf[256];
        sprintf(buf,"%s %i:",typeBuf,s);
        sprintf(sbuf,"SEND %i %s;",strlen(buf),buf);
        if (write(con,sbuf,strlen(sbuf)) != strlen(sbuf))
            return false;
        
        //cout << sbuf << " " << buf << endl;
        //sleep(2);

        curPos = 0;
        while (curPos < s)
        {
            if ((curPos+block) > s)
                block = s - curPos;
            sprintf(sbuf,"SEND %i ",block);
            int l = strlen(sbuf);
            memcpy(sbuf+l,data+curPos,block);
            memcpy(sbuf+l+block,";",1);
            n = write(con,sbuf,l+block+1);
            if (n != l+block+1)
                return false;
            curPos += block;
            //cout << curPos << " of " << s << " bytes transfered block: "  << block << endl;
        }
        //cout << "transfered " << curPos << " bytes of " << s << endl;
    }

    return true;
}

Packet* Connection::recv()
{
    int n,s,curPos;
    char buf[246];
    char hdr[246];
    char* c;
    char* data;
    bool bNeedMoreBytes;
    int type;

    s = n = curPos = 0;
    c = data = 0;
    bNeedMoreBytes = true;
    memset(hdr,0,sizeof(hdr));
    memset(buf,0,sizeof(buf));

    if (excessBytes > 0)
    {
        cout << "copying in excess..." << endl;
        n = excessBytes;
        memcpy(buf,excess,n);
        excessBytes = 0;
    }

    do
    {
        buf[n] = '\0';
        if (n > 0)
        {
            memcpy(hdr+curPos,buf,n);
            curPos+=n;
        }
        c = strstr(hdr,":");
        if (c)
        {
            //cout << "hdr found" << endl;
            break;
        }

    } while ((n = read(con,buf,sizeof(buf)-1)) > 0);

    if (n <= 0)
        return NULL;

    c = strstr(hdr,":");
    if (c)
    {
        *c = 0;
        s = atoi(strstr(hdr," ")+1);
        //cout << "got s: " << s << " hdr: " << hdr << endl;
        excessBytes = curPos-(strlen(hdr)+1);
        if (excessBytes > excessMax)
        {
            //cout << "doubling excess buffer..." << endl;
            excessMax *= 2;
            free(excess);
            excess = (char *)malloc(excessMax);
        }
        //cout << "EXCESS!" << endl;
        memcpy(excess,c+1,excessBytes);
        //cout << "eBytes: " << excessBytes << " excess: " << excess << endl;

        //for (int i=0;i<excessBytes;i++) cout << excess[i] << "-"; cout << endl;
    }

    type = n = curPos = 0;
    if (strstr(hdr,"MSG"))  type = MSG;
    if (strstr(hdr,"IMG"))  type = IMG;
    if (strstr(hdr,"CMD"))  type = CMD;
    if (strstr(hdr,"DATA")) type = DATA;

    if (excessBytes > 0)
    {
        //cout << "loading excess..." << endl;
        n = excessBytes;
        memcpy(buf,excess,n);
        excessBytes = 0;
    }

    //data = (char *)malloc(s);
    //p = new Packet(type,s);
    //data = p->data;

    //packetMutex.lock();
    if (p) delete p;
    p = new Packet(type,s);
    data = p->data;
    //packetMutex.unlock();

    do
    {
        buf[n] = '\0';
        //cout << buf << endl;
        if (n > 0)
        {
            //packetMutex.lock();
            memcpy(data+curPos,buf,n);
            curPos+=n;
            //packetMutex.unlock();
        }
        if (curPos >= s)
            break;

        //cout << "reading into buf..." << endl;
    } while ((n = read(con,buf,sizeof(buf)-1)) > 0);

    if (n <= 0)
        return NULL;

    if (curPos > s)
    {
        //cout << "EXCESS!" << endl;
        excessBytes = curPos - s;
        if (excessBytes > excessMax)
        {
            //cout << "doubling excess buffer..." << endl;
            excessMax *= 2;
            free(excess);
            excess = (char *)malloc(excessMax);
        }

        //packetMutex.lock();
        memcpy(excess,data+s,excessBytes);
        //packetMutex.unlock();
    }
    else
    {
        excessBytes = 0;
        memset(excess,0,excessMax);
    }
   
    //cout << type << " " << s << " " << data << " " << excess << endl;
    //for (int i=0;i<excessBytes;i++) cout << excess[i] << "-"; cout << endl;

    return p;
}

void Connection::close(bool bSendClose)
{
    if (bSendClose)
    {
        char msg[256];
        strcpy(msg,"CLOSE;");
        if (write(con,msg,strlen(msg)) != strlen(msg))
            return;
        cout << "sent: " << msg << endl;
    }
    if (excessMax > 0)
        free(excess);
    
    tcp_close(fdl);
}
