#include "ImgSrv.h"

using namespace std;

void ImgSrv::error(char *msg)
{
    perror(msg);
    exit(0);
}

// Sets up connection
ImgSrv::ImgSrv(char* addr, char* sPort) {
    struct sockaddr_in serv_addr;
    struct hostent *server;

    int port = atoi(sPort);

    sockfd = socket(AF_INET, SOCK_STREAM, 0);
    if (sockfd < 0) 
        error("ERROR opening socket");
    server = gethostbyname(addr);
    if (server == NULL) {
        fprintf(stderr,"ERROR, no such host\n");
        exit(0);
    }
    bzero((char *) &serv_addr, sizeof(serv_addr));
    serv_addr.sin_family = AF_INET;
    bcopy((char *)server->h_addr, 
            (char *)&serv_addr.sin_addr.s_addr,
            server->h_length);
    serv_addr.sin_port = htons(port);
    if (connect(sockfd,(const sockaddr*)&serv_addr,sizeof(serv_addr)) < 0) 
        error("ERROR connecting");
    p = NULL;
}

ImgSrv::~ImgSrv() {
    close(sockfd);
    //cout << "close: " << close(sockfd) << endl;
    //cout << "closing..." << endl;
}

gdImagePtr ImgSrv::getImage(char* msg) {
    int n; // Number of bytes read
    char type; // Type of image (quality)
    char s[4]; // Size of image (in array)
    unsigned int size; // Size of image

    bzero(&type,1);
    bzero(s,4);

    // If I have a message, send it at this time
    if (msg)
        write(sockfd,msg,strlen(msg));
    
    // Get type
    n = read(sockfd,&type,1);
    //cout << "type = " << (int) type << endl;

    // Get size
    usleep(80);
    //cout << "Here1" << endl;
    n = read(sockfd,s,4);
    //cout << "here" << endl;
    size = (unsigned char)s[3] + ((unsigned char)s[2])*256 + ((unsigned char)s[1])*65536 + ((unsigned char)s[0])*16777216;

    //cout << size << endl;
    char buffer[size];
    bzero(buffer,size);

    // Get picture
    usleep(size*20);
    n = read(sockfd,buffer,size);

    if (n < 0) 
        error("ERROR reading from socket");

    return gdImageCreateFromJpegPtr(size,buffer);
}

Packet* ImgSrv::front()
{
    imgMutex.lock();
    return p;
}

void ImgSrv::pop_front()
{
    imgMutex.unlock();
}

void ImgSrv::Setup()
{
    cout << "ImgSrv::Setup" << endl;
    p = new Packet(0,1);
}

void ImgSrv::Execute()
{
    cout << "ImgSrv::Execute" << endl;
    while (1)
    {
        img = getImage("OK");
        int s;
        char* buf = (char *)gdImageJpegPtr(img,&s,50);

        if (imgMutex.trylock())
        {
            cout << "ImgSrv::Execute(): got image" << endl;
            if (p)
                delete p;
            p = new Packet(IMG,s,buf);
            imgMutex.unlock();
        }

        gdImageDestroy(img);
        gdFree(buf);
        Wait();
        //usleep(2000);
        //sleep(1);
    }
}

void ImgSrv::Exit()
{
    cout << "ImgSrv::Exit" << endl;
}
