#include <SDL/SDL.h>
#include <SDL/SDL_joystick.h>
#include <SDL/SDL_keyboard.h>
#include "JoystickManager.h"

using namespace std;

JoystickManager::JoystickManager(int jIndex)
{
    joystickIndex = jIndex;
    p = NULL;
    bReady = false;
    
    if (SDL_Init(SDL_INIT_JOYSTICK) < 0)
        Stop();
    if (joystickIndex != -1)
    {
        joy = SDL_JoystickOpen(joystickIndex);

        // Poll the joystick for a bit to clear out the values.
        while (!bReady)
        {
            SDL_JoystickUpdate();
            bReady = true;

            for (int i=0;i<SDL_JoystickNumButtons(joy);i++)
                if (SDL_JoystickGetButton(joy,i)==1)
                    bReady = false;
            for (int i=0;i<SDL_JoystickNumAxes(joy)-1;i++)
                if (SDL_JoystickGetAxis(joy,i) != 0)
                    bReady = false;
            usleep(500);
        }
    }
    else
    {
        bReady = false;
    }
}

bool JoystickManager::AnyButton()
{
    SDL_JoystickUpdate();
    for (int i=0;i<SDL_JoystickNumButtons(joy);i++)
        if (SDL_JoystickGetButton(joy,i)==1)
            return true;
    return false;
}

Packet* JoystickManager::getPacket()
{
    SDL_JoystickUpdate();
    if (p)
        delete p;
    
    char s[512];
    float maxSpeed = 2.0f;
    float maxROT = 30.0f;
    float maxArmSpeed = 2.0f;

    bool bMoveArm = false;
    float armSpeeds[10];

    for (int i=0;i<10;i++)
        armSpeeds[i] = 0.0f;

    float sens = (float)(SDL_JoystickGetAxis(joy,2)/32000+1)+.5f;
    float xaxis = (float)SDL_JoystickGetAxis(joy,0)/-32000.0f*sens;
    float yaxis = (float)SDL_JoystickGetAxis(joy,1)/-32000.0f*sens;

    for (int i=0;i<5;i++)
    {
        if (SDL_JoystickGetButton(joy,i))
        {
            cout << "Button " << i << endl;
            bMoveArm = true;
            if (i==2)   armSpeeds[i] = -xaxis * maxArmSpeed;
            else        armSpeeds[i] = yaxis * maxArmSpeed;
        }
    }
    if (!bMoveArm)
    {
        //sprintf(s,"MOVE %.2f %.2f 0.0 0.0 0.0 0.0 0.0:",((float)SDL_JoystickGetAxis(joy,1)/-32000.0f)*maxSpeed,((float)SDL_JoystickGetAxis(joy,0)/-32000.0f)*maxROT);
        sprintf(s,"MOVE %.2f %.2f 0.0 0.0 0.0 0.0 0.0:",(yaxis)*maxSpeed,xaxis*maxROT);
        p = new Packet(CMD,strlen(s)+1,s);
    }
    else
    {
        strcpy(s,"MOVE 0.0 0.0");
        for (int i=0;i<5;i++)
        {
            char z[32];
            sprintf(z," %.2f",armSpeeds[i]);
            strcat(s,z);
        }
        strcat(s,":");
        p = new Packet(CMD,strlen(s)+1,s);
    }
    return p;
}

void JoystickManager::Setup()
{
    cout << "JoystickManager::Setup()" << endl;
}

bool JoystickManager::Ready()
{
    return bReady;
}

void JoystickManager::Execute()
{
    cout << "JoystickManager::Execute()" << endl;

    if (joystickIndex==-1)
    {
        /*
        while (1)
        {
            if (jMutex.trylock())
            {
                p = NULL;
                jMutex.unlock();
            }
            sleep(3);
        }*/
    }
    else if (joystickIndex != -1)
    {
    }
    /*
    while (1)
    {
        SDL_JoystickUpdate();
     
        if (SDL_JoystickGetButton(joy,10)==1)
        {
            if (jMutex.trylock())
            {
                if (p)
                {
                    delete p;
                    sprintf(s,"STOP");
                    p = new Packet(CMD,strlen(s)+1,s);
                }
                jMutex.unlock();
                break;
            }
        }

        if (jMutex.trylock())
        {
            if (p)
                delete p;
            float maxSpeed = 3.0f;
            float maxROT = 50.0f;
            float maxArmSpeed = 2.0f;

            bool bMoveArm = false;
            float armSpeeds[10];

            for (int i=0;i<10;i++)
                armSpeeds[i] = 0.0f;

            float xaxis = (float)SDL_JoystickGetAxis(joy,0)/32000.0f * maxArmSpeed;
            float yaxis = (float)SDL_JoystickGetAxis(joy,1)/-32000.0f * maxArmSpeed;

            for (int i=0;i<5;i++)
            {
                if (SDL_JoystickGetButton(joy,i))
                {
                    bMoveArm = true;
                    if (i==2)   armSpeeds[i] = xaxis;
                    else        armSpeeds[i] = yaxis;
                }
            }
            if (!bMoveArm)
            {
                sprintf(s,"MOVE %.2f %.2f 0.0 0.0 0.0 0.0 0.0:",((float)SDL_JoystickGetAxis(joy,1)/-32000.0f)*maxSpeed,((float)SDL_JoystickGetAxis(joy,0)/-32000.0f)*maxROT);
                p = new Packet(CMD,strlen(s)+1,s);
            }
            else
            {
                //sprintf(s,"MOVE 0.0 0.0 %.2f %.2f %.2f %.2f %.2f:",armSpeeds[0]);
                strcpy(s,"MOVE 0.0 0.0");
                for (int i=0;i<5;i++)
                {
                    char z[32];
                    sprintf(z," %.2f",armSpeeds[i]);
                    strcat(s,z);
                }
                strcat(s,":");
                p = new Packet(CMD,strlen(s)+1,s);
            }
            jMutex.unlock();
        }
        usleep(10000);
    }
    */
}

void JoystickManager::Exit()
{
    cout << "JoystickManager::Exit()" << endl;
    SDL_JoystickClose(joy);
    sleep(1);
}
