#include <playerclient.h>
#include "ProxyManager.h"
#include "basic.h"

using namespace std;

ProxyManager::ProxyManager(char* pIP,char* pPort)
{
    pc  = new PlayerClient(pIP,atoi(pPort));
    pp  = new PositionProxy(pc,0,'a');
    lp  = new LaserProxy(pc,0,'a');
    aap = new ActArrayProxy(pc,0,'a');
    lastX = lastY = lastTheta = 0.0f;
    speed = rot = 0.0f;
    p = NULL;
    ofstream os("tmp.log");
    os.close();
    rate = 5000;
    dataCount = 0;
}

void ProxyManager::move(float s,float r,int numLinks,float* lSpeeds)
{
    writeMutex.lock();
    speed = s;
    rot = DTOR(r);

    for (int i=0;i<numLinks;i++)
        linkSpeeds[i] = lSpeeds[i];
   
    writeMutex.unlock();
}

Packet* ProxyManager::front()
{
    readMutex.lock();
    filebuf *pbuf;
    ifstream is;
    int size;
    char* buffer;
    char tmp[256];
    bool bFirst = true;

    is.open("tmp.log");
    pbuf=is.rdbuf();
    size=pbuf->pubseekoff(0,ios::end);
    pbuf->pubseekpos(0);
    
    if (size < 100)
    {
        char s[64];
        sprintf(s,"No Significant Movement...\n");
        p = new Packet(MSG,strlen(s)+1,s);
    }
    else
    {
        buffer = new char[size];
        memset(buffer,0,size);
        do
        {
            if (!(is >> tmp))
                break;
            if (strstr(tmp,"Odometry"))
            {
                if (!bFirst)
                    strcat(buffer,"\n");
                else
                    bFirst = false;
            }
            else if (strstr(tmp,"Laser"))
                strcat(buffer,"\n");
            else
                strcat(buffer," ");
            strcat(buffer,tmp);
        } while (is.good());
        if (p) delete p;
        p = new Packet(DATA,strlen(buffer)+1,buffer);

        is.close();
        ofstream os("tmp.log");
        os.close();
    }
    return p;
}

void ProxyManager::pop_front()
{
    if (p) delete p;
    p = NULL;
    //ofstream os("tmp.log",ios::out);
    //os.close();
    readMutex.unlock();
}

void ProxyManager::Setup()
{
    cout << "ProxyManager::Setup" << endl;
    pc->Read();
    lastX = pp->Xpos();
    lastY = pp->Ypos();
    lastTheta = pp->Theta();
    if      (lastTheta >  M_PI) lastTheta = lastTheta - 2*M_PI;
    else if (lastTheta < -M_PI) lastTheta = lastTheta + 2*M_PI;
    cout << "initial: " << lastX << " " << lastY << " " << lastTheta << endl;
}

void ProxyManager::Execute()
{
    bool bSkip = false;
    cout << "ProxyManager::Execute" << endl;
    int counter = 0;
    time_t prevTime,currTime;
    time(&prevTime);
    rate = 5000;
    while (!bKill)
    {
        if (pc->Read())
            break;

        if (readMutex.trylock())
        {
            double ddist,dangle;
            ddist = sqrt(SQUARE(pp->Xpos()-lastX)+SQUARE(pp->Ypos()-lastY));
            dangle = fabs(pp->Theta()-lastTheta);

            if (ddist > .05 || dangle > .03)
            {
                lastX = pp->Xpos();
                lastY = pp->Ypos();
                lastTheta = pp->Theta();
                
                if      (lastTheta >  M_PI) lastTheta = lastTheta - 2*M_PI;
                else if (lastTheta < -M_PI) lastTheta = lastTheta + 2*M_PI;

                char s[1024];
                sprintf(s,"Odometry %.2f %.2f %.2f \n",lastX,lastY,lastTheta);
                char z[32];
                int count = lp->RangeCount();
                sprintf(z,"Laser %i",count);
                strcat(s,z);
                for (int i=0;i<count;i++)
                {
                    sprintf(z," %.2f",lp->Ranges(i));
                    strcat(s,z);
                }
                ofstream os("tmp.log",ios::out | ios::app);
                os << s << endl;
                os.close();
            }
            /*
            bSkip = false;

            if ((ddist < 0.05) && (dangle < 0.03))
                bSkip = true;
            time(&currTime);
            if ((difftime(currTime,prevTime) > .25f && !bSkip) || size < 300)
            {
                time(&prevTime);
                dataCount++;
                if (size > 10000)
                {
                    int counter=0;
                    char buf[32];
                    char ox[8],oy[8],otheta[8];
                    ifstream is("tmp.log",ios::in);
                    ofstream os("tmp2.log",ios::out);
                    while (is.good())
                    {
                        if (++counter >= 1000) counter = 0;
                        int count = 0;
                        is >> buf;is >> ox;is >> oy;is >> otheta;
                        is >> buf;is >> buf;count = atoi(buf);
                        char laser[181][8];
                        for (int i=0;i<count;i++)
                            is >> laser[i];
                        if (counter%2==0)
                        {
                            os << "Odometry " << ox << " " << oy << " " << otheta << endl;
                            os << "Laser " << count;
                            for (int i=0;i<count;i++)
                                os << " " << laser[i];
                            os << endl;
                        }
                    }
                    os.close();
                    is.close();
                    os.open("tmp.log",ios::out);
                    is.open("tmp2.log",ios::in);
                    os << is.rdbuf();
                    is.close();
                    os.close();
                }
                
            }*/
            readMutex.unlock();
        }

        if (writeMutex.trylock())
        {
            pp->SetSpeed(speed,rot);
            if (aap->GetAccess() != 'e')
                for (int i=0;i<5;i++)
                    aap->Link(i+1,linkSpeeds[i]);
            //aap->Link(currLink,linkSpeeds[currLink]);
            /*
            for (int i=0;i<5;i++)
            {
                if (linkSpeeds[i] != 0.0f)
                    aap->Link(i,linkSpeeds[i]);
            }
            //*/
            writeMutex.unlock();
        }
        else
        {
            pp->SetSpeed(0.0f,0.0f);
            //for (int i=0;i<5;i++)
            //    aap->Link(i,0.0f);
        }
        //usleep(rate);
    }
}

void ProxyManager::Exit()
{
    cout << "ProxyManager::Exit" << endl;
    
    pp->SetSpeed(0.0f,0.0f);
    if (aap->GetAccess() != 'e')
        for (int i=0;i<5;i++)
            aap->Link(i,0.0f);
    
    //aap->Link(1,0.0f);
    //aap->Link(2,0.0f);
}
