#include <iostream>
#include <deque>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>

#include "ConnectionManager.h"
#include "Connection.h"
#include "JoystickManager.h"
#include "Thread.h"

using namespace std;

class R : public Thread
{
    Connection* r;
public:
    R() {
        r = new Connection(UT_IP_ADDR,UT_WCS_PORT,"Talon","7338","localhost");
    }
    ~R() {
        r->close();
        delete r;
    }

    virtual void Setup() {
        cout << "R::Setup " << r->listenNoUT() << endl;
    }
    virtual void Execute() {
        cout << "R::Execute" << endl;
        while (1) {
            cout << "-" << endl;
            Packet* p = r->recv();
            if (p==NULL)
                break;
            else
                cout << p->data << endl;
            //usleep(500);
        }
    }
    virtual void Exit() {
        cout << "R::Exit" << endl;
    }
};

class S : public Thread
{
    Mutex sMutex;
    Connection* s;
    deque<Packet> sBuffer;
public:
    S() {
        s = new Connection(UT_IP_ADDR,UT_WCS_PORT,"BaseStation","7339","localhost");
    }
    ~S() {
        s->close();
        delete s;
    }

    virtual void Setup() {
        cout << "S::Setup " << s->connectNoUT("7338") << endl;

        char z[8000];
        for (int i=0;i<50;i++) {
            //sprintf(z,"hihi! %i",i);
            memset(z,'A',7999);
            memset(z+7999,'\0',1);
            cout << strlen(z) << endl;
            Packet p(MSG,strlen(z)+1,z);
            sBuffer.push_back(p);
            //s->send(&p,false);
        }

    }
    virtual void Execute() {
        cout << "S::Execute" << endl;
        while (sBuffer.size() > 0) {
            if (sMutex.trylock()) {
                s->send(&sBuffer.front(),false);
                sBuffer.pop_front();
                sMutex.unlock();
            }
            usleep(50000);
        }

    }
    virtual void Exit() {
        cout << "S::Exit" << endl;
    }
};

class Monitor : public Thread
{
    int* counter;
    SignalMutex* m;
public:
    Monitor(int* c,SignalMutex* mut) {
        counter = c;
        m = mut;
    }
    virtual void Setup(){
        cout << "Monitor::Setup" << endl;
    }
    virtual void Execute() {
        cout << "Monitor::Execute" << endl;
        while (1)
        {
            m->lock();
            cout << *counter << endl;
            m->unlock();
            m->signal();
            sleep(1);
        }
    }
    virtual void Exit() {
        cout << "Monitor::Exit" << endl;
    }
};

class Updater : public Thread
{
    int* counter;
    SignalMutex* m;
public:
    Updater(int* c,SignalMutex* mut) {
        counter = c;
        m = mut;
    }
    virtual void Setup() {
        cout << "Updater::Setup" << endl;
    }
    virtual void Execute() {
        cout << "Updater::Execute" << endl;
        while (1)
        {
            if (m->trylock())
            {
                m->wait();
                int before = *counter;
                *counter = *counter + 2;
                int after = *counter;
                if ((before+2) != after)
                    cout << "inconsistant" << endl;
                m->unlock();
            }
            usleep(rand()%50000);
        }
    }
    virtual void Exit() {
        cout << "Updater::Exit" << endl;
    }
};

class App : public Thread
{
    SignalMutex mut;
    int counter;

    S* s;
    R* r;
    Monitor* m;
    Updater* u[1000];
    JoystickManager* jm;
    int numUpdaters;
public:
    App() {
        r = new R;
        s = new S;
        //numUpdaters = 1000;
        //m = new Monitor(&counter,&mut);
        //for (int i=0;i<numUpdaters;i++)
        //    u[i] = new Updater(&counter,&mut);
        //jm = new JoystickManager(1);
    }
    virtual void Setup() {
        cout << "App::Setup" << endl;
        //m->Start();
        //for (int i=0;i<numUpdaters;i++)
        //    u[i]->Start();
        r->Start();
        //usleep(500000);
        s->Start();
        //jm->Start();
        //while (!jm->Ready()) ;
    }
    virtual void Execute() {
        cout << "App::Execute" << endl;
        sleep(20);
    }
    virtual void Exit() {
        cout << "App::Exit" << endl;
    }
};

int main(int argc,char** argv)
{
    App* a = new App;

    a->Start();

    Thread::Wait(a);
    
    a->Stop();

    return 0;
}
